renoise.tool():add_keybinding {
  name = "Global:Tools:Set Meta Values",
  invoke = function()main()end
}

renoise.tool():add_menu_entry {
  name = "Pattern Editor:Selection:Set Meta Values",
  invoke = function()main()end
}

renoise.tool():add_menu_entry {
  name = "Pattern Editor:Set Meta Values in Selection...",
  invoke = function()main()end
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:Set Meta Values",
  invoke = function()main()end
  
}


local my_dialog = nil 

local hex_table_vol_pan = {"Clear Selection"}
local hex_table_delay   = {"Clear Selection"}

local allowed_chars = { --n.b. case insensitive
  "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", 
  "a", "b", "c", "d", "e", "f"
}

--populate vol,pan table
for i = 0, 128 do -- 00h .. 80h
  hex_table_vol_pan[i + 2] = string.format("%X", i)
  --add leading zeros to single digits
  if #hex_table_vol_pan[i + 2] < 2 then
    hex_table_vol_pan[i + 2] = "0"..hex_table_vol_pan[i + 2]
  end
end

--populate delay table
for i = 0, 255 do -- 00h .. FFh
  hex_table_delay[i + 2] = string.format("%X",i)
   --add leading zeros to single digits
  if #hex_table_delay[i + 2] < 2 then
    hex_table_delay[i + 2] = "0"..hex_table_delay[i + 2]
  end
end

local dont_constrain = false
local lock_slider = false
local holder = 1



-------------------------------------------------------
--main
-------------------------------------------------------

function main()
  local string_holder = {}
  local pattern_iter = renoise.song().pattern_iterator
  local pattern_index = renoise.song().selected_pattern_index
  
  local vb = renoise.ViewBuilder()
  local NOTE_OFF = 120
  
  -- only allows one dialog instance
  if (my_dialog and my_dialog.visible) then 
    my_dialog:show()
    return
  end


  -------------------------------------------------------
  --local functions
  -------------------------------------------------------
  --closer fn
  local function closer(d)
  
    if d and d.visible then             
        d:close()
    end
  end
  
  --ternary operator
  local function iif(condition, if_true, if_false)
    if condition then return if_true else return if_false end
  end
  
  --split a string by a delimiter
  function string:split(delimiter_regex)
    local tbl = {}
    for v in string.gmatch(self, delimiter_regex) do
      table.insert(tbl, v)
    end
    return tbl
  end
  
  --Get no of tracks
  local function get_number_of_tracks() 
  
  local track_table = renoise.song().tracks
  local note_tracks = 0  --incrementor
  local NOTE_TRACK = 1
  
    for _,track in ipairs(track_table) do
     
      if track.type == NOTE_TRACK then
        note_tracks = note_tracks + 1
      end    
    end   
   return note_tracks
  end
  
  local function slider_to_numeric(value)
    if value == nil then
      return value
    end
    return value * iif(vb.views["switch"].value < 3, 128, 255)
  end
  
  local function numeric_to_hex(value)
    if value == nil then
      return value
    end
    local str = string.sub(bit.tohex(value), -2):upper()
    str = str:gsub("^0+", "")
    if #str == 0 then
      str = "0"
    end
    return str
  end
  
  local function slider_to_hex(value)
    return numeric_to_hex(slider_to_numeric(value))
  end
  
  local function numeric_to_slider(value)
    local output
    if value == nil then
      output = value
    else
      output = value / iif(vb.views["switch"].value < 3, 128, 255) --scale differently depending on current type
    end
    return output
  end
  
  local function hex_to_numeric(str)
    return tonumber(str, 16)
  end
  
  local function hex_to_slider(str)
    return numeric_to_slider(hex_to_numeric(str))
  end
  
  --Update the slider
  local function update_slider()
    if lock_slider then
      return
    end
    local numeric_value = hex_to_numeric(table.concat(string_holder)) --nil if not hex
    local new_slider_value = numeric_to_slider(numeric_value) --nil if not numeric
    if vb.views['slider'].value ~= new_slider_value then
      if new_slider_value == nil then
        lock_slider = true
        vb.views['slider'].value = 0
        lock_slider = false
      elseif new_slider_value >= 0 and new_slider_value <= 1 then
        lock_slider = true
        vb.views['slider'].value = new_slider_value
        lock_slider = false
      end
    end
  end
  
  --Update the text reflection of that
  local function update_text()
    vb.views["text"].text = table.concat(string_holder) --update gui
  end
  
  --Constrain the value based on the current type
  local function constrain_value()
    if dont_constrain then
      return
    end
    local potential_slider_value = numeric_to_slider(hex_to_numeric(table.concat(string_holder)))
    if potential_slider_value ~= nil and potential_slider_value > 1 then
      potential_slider_value = 1
      string_holder = slider_to_hex(potential_slider_value):split(".")
    end
  end
  
  --Update the UI without setting the value in one shot
  local function update_ui()
    update_slider()
    update_text()
  
    --vars
    local typed_number_string = table.concat(string_holder)
    
    --map typed hex string back to holder
    holder = hex_to_numeric(typed_number_string) --nil if not valid hex
      
    if holder then
      --for vol & pan, maps holder value < 0 to clear value (i.e. 255)
      if holder < 0 and vb.views["switch"].value < 3 then
        holder = -1
      end
      if vb.views["switch"].value < 3 and holder > 128 then --reduced maximum if vol, pan chosen after delay
        holder = 128
      end
    else
      holder = -1
    end
    
   -- vb.views["popup_hex"].value = holder + 2 --set hex dropdown to correct hex number
    
    return holder
  end
  
  --Set the value in one shot
  local function set_value(value)
    if value == nil then
      string_holder = {}
    else
      string_holder = tostring(value):upper():split(".")
    end
    update_ui()
  end
  
  --Set the value from the slider
  local function set_value_from_slider(value)
    set_value(slider_to_hex(value))
  end
  
  local function do_the_job()
    for track_index = 1, get_number_of_tracks() do
      for pos, line in pattern_iter:note_columns_in_pattern_track(pattern_index, track_index) do
        if line.note_value < NOTE_OFF or --vol and pan don't affect noteoffs, but delay does
           line.note_value == NOTE_OFF and vb.views["switch"].value == 3 or
           line.note_value == NOTE_OFF and vb.views["switch"].value == 1 and holder < 0 then
          if line.is_selected  then
            --set vol, pan, dly to all subsequent notes 
              
            if vb.views["switch"].value == 1 then --vol
              if not renoise.song().tracks[track_index].volume_column_visible then
                renoise.song().tracks[track_index].panning_volume_visible = true
              end
              if holder >= 0 then
                line.volume_value = holder
              else
                line.volume_string = ".." --clear
              end
            end
            if vb.views["switch"].value == 2 then --pan
              if not renoise.song().tracks[track_index].panning_column_visible then
                renoise.song().tracks[track_index].panning_column_visible = true
              end
              if holder >= 0 then
                line.panning_value = holder
              else
                line.panning_string = ".." --clear
              end
            end
            if vb.views["switch"].value == 3 then --delay
              if not renoise.song().tracks[track_index].delay_column_visible then
                renoise.song().tracks[track_index].delay_column_visible = true
              end
              if holder >= 0 then
                line.delay_value = holder
              else
                line.delay_string = ".." --clear
              end
            end  
                   
          end
        end
      end
    end
  
    --clear
    string_holder = {}
    
    --vb.views["text"].value = ""
  
    closer(my_dialog)
  end
  
  
  
  --------------------------------------------------------
  --GUI
  --------------------------------------------------------
  local dialog_metric_width = 150;
  
  local dialog_content = 
  
  vb:vertical_aligner {
    margin = 8,
    mode = "center",
    spacing = 8,
  
    vb:row {
      style = "group",
      vb:column {
        --see what is typed
        vb:text {
          text = "Clear Range",
          id = "text",
          align = "center",
          font = "big",
          width = dialog_metric_width,
        },
        vb:row {
          vb:column {
            margin = 3,
            -- reset to Clear button
            vb:button {
              bitmap = "Clear.bmp",
              width = 18,
              height = 18,
              notifier = function(value)
                set_value(nil)
                vb.views["text"].text = "Clear Range"
              end
              }
          },
          -- slider control
          vb:minislider {
            id = 'slider',
            value = 0,
            width = dialog_metric_width - 24,
            height = 24,
            notifier = function(value)
              if lock_slider then
                return
              end
              lock_slider = true
              set_value_from_slider(value)
              lock_slider = false
            end
          }
        }
      }
    },
    --hex value dropdown
   --[[vb:popup {
      id = "popup_hex",
      width = dialog_metric_width,
      value = 1,
      items = hex_table_vol_pan, --default dropdown items
    },--]]
    --chosen type switch                    
    vb:switch {
      id = "switch",
      width = dialog_metric_width,
      value = 1,
      items = { "V\nVelocity","P\nPan","L\nDelay"},
      height = 40,
      notifier = function(number) --change dropdown items on type switch
        if number < 3 then
         -- vb.views["popup_hex"].items = hex_table_vol_pan
        else
         -- vb.views["popup_hex"].items = hex_table_delay
        end
        constrain_value()
        update_ui()        
      end
    },
    vb:button {
      width = dialog_metric_width,
      text = 'Apply',
      notifier = function()
        do_the_job()
      end
    }
  }
  
  
  
  
  
  ---------------------------------------------------------
  --keyhandler
  ---------------------------------------------------------
  --oprint(key)
  
  --[character] =>  w
  --[modifiers] =>  
  --[name] =>  w
  --[note] =>  14
  --[repeated] =>  false
  
  
  
  local function my_keyhandler_func(dialog, key)
  
    pattern_index = renoise.song().selected_pattern_index
    
    --esc to close
    if key.modifiers == "" and key.name == "esc" then
      closer(my_dialog)
      return
    end
    
    --convert characters / names to lowercase
    if key.character then
      key.character = key.character:lower()
    end
    if key.name then
      key.name = key.name:lower()
    end
    
    --choose popup value (vol, pan, delay)
    if key.character then
      if key.character == "v" then
        vb.views["switch"].value = 1 --vol
        return
      end
    
      if key.character == "p" then
        vb.views["switch"].value = 2 --pan
        return
      end
    
      if key.character == "l" then
        vb.views["switch"].value = 3 --delay
        return
      end
    end
     
    --filter characters
    local allowed_char = table.find(allowed_chars, key.name:lower())
    
    --capture string in table
    if key.character and allowed_char then 
      if #string_holder < 2 then --two characters only
        table.insert(string_holder,key.character) --add character to string_holder table
      else --else start again
        vb.views["text"].text = "Clear Selection"
        vb.views["text"].text = key.character
        string_holder = {key.character}
        return
      end
      dont_constrain = true
    end
    
    -- backspace key
    if key.name == "back" then 
      table.remove(string_holder) --remove last character
    end
    
    -- transform keys
    local transform_keys = {
      'prior', 'next', '[', ']', 'up', 'down', 'left', 'right', 'home', 'end', '\\'
    }
    local is_transform_key = table.find(transform_keys, key.name)
    
    local function reset_numeric_using_type(numeric)
      if vb.views["switch"].value == 1 then --vol
        return 1
      elseif vb.views["switch"].value == 2 then --pan
        return 0.5
      elseif vb.views["switch"].value == 3 then --delay
        return 0
      end
      return nil
    end
    
    if is_transform_key then
      local numeric = hex_to_slider(table.concat(string_holder))
      if numeric == nil then
        numeric = reset_numeric_using_type(numeric)
        if vb.views["switch"].value == 1 then --vol override reset -- this seems more natural
          numeric = 0
        end
      end
      if key.name == "]" then
        if vb.views["switch"].value == 2 then --special "angle-like" behaviour for panning
          if numeric < 1 then
            local angle = math.acos( 1 - numeric * 2 ) / math.pi * 360
            angle = angle + 30
            numeric = - math.cos( angle / 360 * math.pi ) / 2 + 0.5
          end
        elseif vb.views["switch"].value == 3 then --special behaviour for delay -- decrease x in the series 1/x, x e Z
          if numeric == nil then
            numeric = 1 / 256
          elseif numeric < 12 / 256 then --this is when integers become too imprecise for the other equation to work
            numeric = numeric + 1 / 256
          else 
            numeric = 1 / ( 1 / numeric - 1) 
          end
        else
          if numeric == 0 then
            numeric = 1 / 128
          else
            numeric = numeric * 2
          end
        end
      end
      if key.name == "[" then
        if vb.views["switch"].value == 2 then --special "angle-like" behaviour for panning
          if numeric > 0 then
            local angle = math.acos( 1 - numeric * 2 ) / math.pi * 360
            angle = angle - 30
            numeric = - math.cos( angle / 360 * math.pi ) / 2 + 0.5
          end
        elseif vb.views["switch"].value == 3 then --special behaviour for delay -- increase x in the series 1/x, x e Z
          if numeric <= 1 / 256 then
            numeric = 0
          elseif numeric < 12 / 256 then --this is when integers become too imprecise for the other equation to work
            numeric = numeric - 1 / 256
          else
            numeric = 1 / ( 1 / numeric + 1)
          end
        else
          if numeric <= 1 / 128 then
            numeric = 0
          else 
            numeric = numeric / 2
          end
        end
      end
      if key.name == "prior" then
        numeric = numeric - 1/8
      end
      if key.name == "next" then
        numeric = numeric + 1/8
      end
      if key.name == "up" then
        numeric = numeric + 1/16
      end
      if key.name == "down" then
        numeric = numeric - 1/16
      end
      if key.name == "left" then
        numeric = numeric - 1/255.9
      end
      if key.name == "right" then
        numeric = numeric + 1/256
      end
      if key.name == "home" then
        numeric = 0
      end
      if key.name == "end" then
        numeric = 1
      end
      if key.name == "\\" then --this is the 'neutral value' key
        numeric = reset_numeric_using_type(numeric)
      end
      
      if numeric ~= nil then
        if numeric > 1 then
          numeric = 1 --allowing it to go over does weird stuff
        end
        if numeric < 0 then
          numeric = 0 --allowing it to go over does weird stuff
        end
      end
      
      string_holder = slider_to_hex(numeric):split(".")
    end
    
    constrain_value()
    update_ui()
    
    dont_constrain = false
    
    --print(holder)
    --execute
    if key.name == "return" then
      do_the_job()
    end
  
  end --key handler
  
  
  ------------------------------------------------------------------------------------------
  ------------------------------------------------------------------------------------------
  --Script dialog
  my_dialog = renoise.app():show_custom_dialog(
      "Set Meta Value", dialog_content,my_keyhandler_func)
  
end --main






